@@ -37,7 +37,9 @@
 
 #include <zlib.h>
 
-extern MDFNGI EmulatedSS;
+using namespace Mednafen;
+
+MDFN_HIDE extern MDFNGI EmulatedSS;
 
 #include "ss.h"
 #include "sound.h"
@@ -55,10 +57,11 @@ namespace MDFN_IEN_SS
 
 static sscpu_timestamp_t MidSync(const sscpu_timestamp_t timestamp);
 
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
 uint32 ss_dbg_mask;
 static std::bitset<0x200> BWMIgnoreAddr[2]; // 0=read, 1=write
 #endif
+uint32 ss_horrible_hacks;
 
 static bool NeedEmuICache;
 static const uint8 BRAM_Init_Data[0x10] = { 0x42, 0x61, 0x63, 0x6b, 0x55, 0x70, 0x52, 0x61, 0x6d, 0x20, 0x46, 0x6f, 0x72, 0x6d, 0x61, 0x74 };
@@ -94,13 +97,14 @@ static int64 CartNV_SaveDelay;
 static uintptr_t SH7095_FastMap[1U << (32 - SH7095_EXT_MAP_GRAN_BITS)];
 
 int32 SH7095_mem_timestamp;
-uint32 SH7095_BusLock;
+static uint32 SH7095_BusLock;
 static uint32 SH7095_DB;
 #include "scu.inc"
 
 #include "debug.inc"
 
 static sha256_digest BIOS_SHA256;	// SHA-256 hash of the currently-loaded BIOS; used for save state sanity checks.
+static int ActiveCartType;		// Used in save states.
 static std::vector<CDInterface*> *cdifs = NULL;
 static std::bitset<1U << (27 - SH7095_EXT_MAP_GRAN_BITS)> FMIsWriteable;
 
@@ -634,14 +638,14 @@ static INLINE bool EventHandler(const sscpu_timestamp_t timestamp)
 
  while(timestamp >= (e = events[SS_EVENT__SYNFIRST].next)->event_time)	// If Running = 0, EventHandler() may be called even if there isn't an event per-se, so while() instead of do { ... } while
  {
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
   const sscpu_timestamp_t etime = e->event_time;
 #endif
   sscpu_timestamp_t nt;
 
   nt = e->event_handler(e->event_time);
 
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
   if(MDFN_UNLIKELY(nt <= etime))
   {
    fprintf(stderr, "which=%d event_time=%d nt=%d timestamp=%d\n", (int)(e - events), etime, nt, timestamp);
@@ -801,7 +805,7 @@ static void Emulate(EmulateSpecStruct* espec_arg)
  int32 end_ts;
 
  espec = espec_arg;
- AllowMidSync = MDFN_GetSettingB("ss.midsync");
+ AllowMidSync = true;
  MDFNGameInfo->mouse_sensitivity = MDFN_GetSettingF("ss.input.mouse_sensitivity");
 
  cur_clock_div = SMPC_StartFrame(espec);
@@ -1067,7 +1071,7 @@ static INLINE bool DetectRegion(unsigned* const region)
 
  return false;
 }
-
+#if 0
 static MDFN_COLD bool DetectRegionByFN(const std::string& fn, unsigned* const region)
 {
  std::string ss = fn;
@@ -1134,12 +1138,12 @@ static MDFN_COLD bool DetectRegionByFN(const std::string& fn, unsigned* const re
 
  return false;
 }
-
-static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned cart_type, const unsigned smpc_area, Stream* dbg_cart_rom_stream)
+#endif
+static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned horrible_hacks, const unsigned cart_type, const unsigned smpc_area, Stream* dbg_cart_rom_stream)
 {
  const char* cart_rom_path_sname = nullptr;
 
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
  ss_dbg_mask = SS_DBG_ERROR;
  {
   std::vector<uint64> dms = MDFN_GetSettingMultiUI("ss.dbg_mask");
@@ -1188,6 +1192,9 @@ static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned
   MDFN_printf(_("CPU Cache Emulation Mode: %s\n"), cem);
  }
  //
+ if(horrible_hacks)
+  MDFN_printf(_("Horrible hacks: 0x%08x\n"), horrible_hacks);
+ //
  {
   MDFN_printf(_("Region: 0x%01x\n"), smpc_area);
   const struct
@@ -1228,6 +1235,10 @@ static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned
   CPU[c].Init(cpucache_emumode == CPUCACHE_EMUMODE_DATA_CB);
   CPU[c].SetMD5((bool)c);
  }
+ SH7095_mem_timestamp = 0;
+ SH7095_DB = 0;
+
+ ss_horrible_hacks = horrible_hacks;
 
  //
  // Initialize backup memory.
@@ -1255,6 +1266,7 @@ static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned
   }
 
   CART_Init(cart_type, cart_rom_stream ? cart_rom_stream.get() : dbg_cart_rom_stream);
+  ActiveCartType = cart_type;
  }
  //
  //
@@ -1264,6 +1276,7 @@ static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned
  const char* biospath_sname;
  int sls = MDFN_GetSettingI(PAL ? "ss.slstartp" : "ss.slstart");
  int sle = MDFN_GetSettingI(PAL ? "ss.slendp" : "ss.slend");
+ const uint64 vdp2_affinity = MDFN_GetSettingUI("ss.affinity.vdp2");
 
  if(PAL)
  {
@@ -1334,7 +1347,7 @@ static void MDFN_COLD InitCommon(const unsigned cpucache_emumode, const unsigned
  SCU_Init();
  SMPC_Init(smpc_area, MasterClock);
  VDP1::Init();
- VDP2::Init(PAL);
+ VDP2::Init(PAL, vdp2_affinity);
  CDB_Init();
  SOUND_Init();
 
@@ -1506,11 +1519,20 @@ static MDFN_COLD void Load(GameFile* gf)
 
    static std::vector<CDInterface*> CDInterfaces;
    CDInterfaces.clear();
-   CDInterfaces.push_back(CDInterface::Open(&NVFS, MDFN_GetSettingS("ss.dbg_exe_cdpath"), false));
+   CDInterfaces.push_back(CDInterface::Open(&NVFS, MDFN_GetSettingS("ss.dbg_exe_cdpath"), false, MDFN_GetSettingUI("affinity.cd")));
    cdifs = &CDInterfaces;
   }
+  //
+  //
+  uint32 horrible_hacks = 0;
+  {
+   std::vector<uint64> dhhs = MDFN_GetSettingMultiUI("ss.dbg_exe_hh");
 
-  InitCommon(CPUCACHE_EMUMODE_DATA, CART_MDFN_DEBUG, MDFN_GetSettingUI("ss.region_default"), gf->stream);
+   for(uint64 dhhse : dhhs)
+    horrible_hacks |= dhhse;
+  }
+
+  InitCommon(MDFN_GetSettingUI("ss.dbg_exe_cem"), horrible_hacks, CART_MDFN_DEBUG, MDFN_GetSettingUI("ss.region_default"), gf->stream);
  }
  catch(...)
  {
@@ -1604,6 +1626,7 @@ static MDFN_COLD void LoadCD(std::vector<CDInterface*>* CDInterfaces)
   unsigned region;
   int cart_type;
   unsigned cpucache_emumode;
+  unsigned horrible_hacks;
   uint8 fd_id[16];
   char sgid[16 + 1] = { 0 };
   cdifs = CDInterfaces;
@@ -1617,6 +1640,7 @@ static MDFN_COLD void LoadCD(std::vector<CDInterface*>* CDInterfaces)
 
   DetectRegion(&region);
   DB_Lookup(nullptr, sgid, fd_id, &region, &cart_type, &cpucache_emumode);
+  horrible_hacks = DB_LookupHH(sgid, fd_id);
   //
   if(!MDFN_GetSettingB("ss.region_autodetect"))
    region = region_default;
@@ -1631,7 +1655,7 @@ static MDFN_COLD void LoadCD(std::vector<CDInterface*>* CDInterfaces)
 
    // TODO: auth ID calc
 
-  InitCommon(cpucache_emumode, cart_type, region, nullptr);
+  InitCommon(cpucache_emumode, horrible_hacks, cart_type, region, nullptr);
  }
  catch(...)
  {
@@ -1642,7 +1666,7 @@ static MDFN_COLD void LoadCD(std::vector<CDInterface*>* CDInterfaces)
 
 static MDFN_COLD void CloseGame(void)
 {
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
  VDP1::MakeDump("/tmp/vdp1_dump.h");
  VDP2::MakeDump("/tmp/vdp2_dump.h");
 #endif
@@ -1854,19 +1878,38 @@ static MDFN_COLD void StateAction(StateMem* sm, const unsigned load, const bool
  if(!data_only)
  {
   sha256_digest sr_dig = BIOS_SHA256;
+  int cart_type = ActiveCartType;
 
   SFORMAT SRDStateRegs[] = 
   {
    SFPTR8(sr_dig.data(), sr_dig.size()),
+   SFVAR(cart_type),
    SFEND
   };
 
-  MDFNSS_StateAction(sm, load, data_only, SRDStateRegs, "BIOS_HASH", true);
+  MDFNSS_StateAction(sm, load, data_only, SRDStateRegs, "BIOS_HASH");
 
-  if(load && sr_dig != BIOS_SHA256)
-   throw MDFN_Error(0, _("BIOS hash mismatch(save state created under a different BIOS)!"));
+  if(load)
+  {
+   if(sr_dig != BIOS_SHA256)
+    throw MDFN_Error(0, _("BIOS hash mismatch(save state created under a different BIOS)!"));
+/*
+   if(load < 0x00102300)
+   {
+    SFORMAT DummyStateRegs[] = { SFEND };
+
+    if(MDFNSS_StateAction(sm, load, data_only, DummyStateRegs, "CART_BACKUP", true))
+     cart_type = CART_BACKUP_MEM;
+   }
+*/
+   if(cart_type != ActiveCartType)
+    throw MDFN_Error(0, _("Cart type mismatch(save state created with a different cart)!"));
+  }
  }
-
+ //
+ //
+ //
+ bool RecordedNeedEmuICache = load ? false : NeedEmuICache;
  EventsPacker ep;
  ep.Save();
 
@@ -1887,6 +1930,8 @@ static MDFN_COLD void StateAction(StateMem* sm, const unsigned load, const bool
   SFVAR(WorkRAMH),
   SFVAR(BackupRAM),
 
+  SFVAR(RecordedNeedEmuICache),
+
   SFEND
  };
 
@@ -1913,6 +1958,14 @@ static MDFN_COLD void StateAction(StateMem* sm, const unsigned load, const bool
    printf("Bad state events data.");
    InitEvents();
   }
+
+  if(NeedEmuICache && !RecordedNeedEmuICache)
+  {
+   //printf("NeedEmuICache=%d, RecordedNeedEmuICache=%d\n", NeedEmuICache, RecordedNeedEmuICache);
+
+   for(size_t i = 0; i < 2; i++)
+    CPU[i].FixupICacheModeState();	// Only call it after all RAM has been loaded from the save state.
+  }
  }
 }
 
@@ -2005,7 +2058,7 @@ static const MDFNSetting_EnumList CartAD_List[] =
  { NULL, 0 },
 };
 
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
 static const MDFNSetting_EnumList DBGMask_List[] =
 {
  { "0",		0								},
@@ -2049,6 +2102,29 @@ static const MDFNSetting_EnumList DBGMask_List[] =
 };
 #endif
 
+static const MDFNSetting_EnumList CEM_List[] =
+{
+ { "data_cb",	CPUCACHE_EMUMODE_DATA_CB,	gettext_noop("Data only, with high-level bypass") },
+ { "data",	CPUCACHE_EMUMODE_DATA, 		gettext_noop("Data only") },
+ { "full",	CPUCACHE_EMUMODE_FULL,		gettext_noop("Full") },
+
+ { NULL, 0 },
+};
+
+static const MDFNSetting_EnumList HH_List[] =
+{
+ { "0",			0								},
+ { "none",		0,				gettext_noop("None")		},
+
+ { "nosh2dmaline106",	HORRIBLEHACK_NOSH2DMALINE106,	gettext_noop("nosh2dmaline106") },
+ { "nosh2dmapenalty",	HORRIBLEHACK_NOSH2DMAPENALTY,	gettext_noop("nosh2dmapenalty")	},
+ { "vdp1vram5000fix",	HORRIBLEHACK_VDP1VRAM5000FIX,	gettext_noop("vdp1vram5000fix")	},
+ { "vdp1rwdrawslowdown",HORRIBLEHACK_VDP1RWDRAWSLOWDOWN,gettext_noop("vdp1rwdrawslowdown") },
+ { "vdp1instant",	HORRIBLEHACK_VDP1INSTANT,	gettext_noop("vdp1instant") },
+
+ { NULL, 0 },
+};
+
 static const MDFNSetting SSSettings[] =
 {
  { "ss.bios_jp", MDFNSF_EMU_STATE | MDFNSF_CAT_PATH, gettext_noop("Path to the Japan ROM BIOS"), NULL, MDFNST_STRING, "sega_101.bin" },
@@ -2104,12 +2180,15 @@ static const MDFNSetting SSSettings[] =
  { "ss.slstartp", MDFNSF_NOFLAGS, gettext_noop("First displayed scanline in PAL mode."), NULL, MDFNST_INT, "0", "-16", "271" },
  { "ss.slendp", MDFNSF_NOFLAGS, gettext_noop("Last displayed scanline in PAL mode."), NULL, MDFNST_INT, "255", "-16", "271" },
 
- { "ss.midsync", MDFNSF_NOFLAGS, gettext_noop("Enable mid-frame synchronization."), gettext_noop("Mid-frame synchronization can reduce input latency, but it will increase CPU requirements."), MDFNST_BOOL, "0" },
+ { "ss.affinity.vdp2", MDFNSF_NOFLAGS, gettext_noop("VDP2 rendering thread CPU affinity mask."), gettext_noop("Set to 0 to disable changing affinity."), MDFNST_UINT, "0", "0x0000000000000000", "0xFFFFFFFFFFFFFFFF" },
 
-#ifdef MDFN_SS_DEV_BUILD
+#ifdef MDFN_ENABLE_DEV_BUILD
  { "ss.dbg_mask", MDFNSF_SUPPRESS_DOC, gettext_noop("Debug printf mask."), NULL, MDFNST_MULTI_ENUM, "none", NULL, NULL, NULL, NULL, DBGMask_List },
 #endif
+
  { "ss.dbg_exe_cdpath", MDFNSF_SUPPRESS_DOC | MDFNSF_CAT_PATH, gettext_noop("CD image to use with bootable cart ROM image loading."), NULL, MDFNST_STRING, "" },
+ { "ss.dbg_exe_cem", MDFNSF_SUPPRESS_DOC | MDFNSF_NONPERSISTENT, gettext_noop("Cache emulation mode to use with bootable cart ROM image loading."), NULL, MDFNST_ENUM, "data", NULL, NULL, NULL, NULL, CEM_List },
+ { "ss.dbg_exe_hh", MDFNSF_SUPPRESS_DOC | MDFNSF_NONPERSISTENT, gettext_noop("Horrible hacks to use with bootable cart ROM image loading."), NULL, MDFNST_MULTI_ENUM, "none", NULL, NULL, NULL, NULL, HH_List },
 
  { NULL },
 };
@@ -2143,6 +2222,7 @@ MDFNGI EmulatedSS =
  NULL,
  #endif
  SMPC_PortInfo,
+ DB_GetInternalDB,
  Load,
  TestMagic,
  LoadCD,
